import os
import json
import shutil
from pydantic.dataclasses import dataclass
from rid_lib.core import RID
from .manifest import Manifest
from .utils import b64_encode, b64_decode, JSONSerializable


@dataclass
class CacheBundle(JSONSerializable):
    """Object representing an individual RID cache entry.

    A container object for the cached data associated with an RID. It is 
    returned by the read function of Cache. It represents a "bundle" of 
    the manifest and (optionally) contents of an RID object.
    """
    manifest: Manifest
    contents: dict | None = None


class Cache:
    def __init__(self, directory_path: str):
        self.directory_path = directory_path
        
    def file_path_to(self, rid: RID) -> str:
        encoded_rid_str = b64_encode(str(rid))
        return f"{self.directory_path}/{encoded_rid_str}.json"

    def write(self, rid: RID, cache_bundle: CacheBundle) -> CacheBundle:
        """Writes contents to cache with an autogenerated manifest, returns a CacheBundle."""
        if cache_bundle.manifest.rid != rid:
            raise ValueError("The provided CacheBundle's Manifest RID must be equivalent to the 'rid' param.")

        if not os.path.exists(self.directory_path):
            os.makedirs(self.directory_path)
            
        with open(self.file_path_to(rid), "w") as f:
            json.dump(cache_bundle.to_json(), f, indent=2)

        return cache_bundle
    
    def bundle_and_write(self, rid: RID, data: dict) -> CacheBundle:
        return self.write(
            rid,
            CacheBundle(
                Manifest.generate(rid, data),
                data
            )
        )
    
    def exists(self, rid: RID) -> bool:
        return os.path.exists(
            self.file_path_to(rid)
        )

    def read(self, rid: RID) -> CacheBundle | None:
        """Reads and returns CacheEntry from RID cache."""
        try:
            with open(self.file_path_to(rid), "r") as f:
                return CacheBundle.from_json(json.load(f))
        except FileNotFoundError:
            return None
        
    def read_all_rids(self) -> list[RID]:
        rids = []
        for filename in os.listdir(self.directory_path):
            encoded_rid_str = filename.split(".")[0]
            rid_str = b64_decode(encoded_rid_str)
            rid = RID.from_string(rid_str, allow_prov_ctx=True)
            rids.append(rid)
            
        return rids
                
    def delete(self, rid: RID) -> None:
        """Deletes cache bundle."""
        try:
            os.remove(self.file_path_to(rid))
        except FileNotFoundError:
            return

    def drop(self) -> None:
        """Deletes all cache bundles."""
        try:
            shutil.rmtree(self.directory_path)
        except FileNotFoundError:
            return

